import factory
from factory.fuzzy import FuzzyDecimal
from wbcore.contrib.directory.factories import CompanyFactory, EmailContactFactory

from wbaccounting.models import EntryAccountingInformation


class EntryAccountingInformationFactory(factory.django.DjangoModelFactory):
    class Meta:  # type: ignore
        model = EntryAccountingInformation

    entry = factory.SubFactory("wbcore.contrib.directory.factories.EntryFactory")
    tax_id = factory.Faker("text", max_nb_chars=64)
    vat = FuzzyDecimal(0, 0.9, 4)
    send_mail = factory.Faker("pybool")
    counterparty_is_private = False

    email_body = factory.Faker("paragraph")

    @factory.post_generation
    def post(self, create, extracted, **kwargs):
        if isinstance(self, EntryAccountingInformationFactory) and not create:
            self.email_to.add(EmailContactFactory.create())
            self.email_cc_add(EmailContactFactory.create())
            self.email_bcc.add(EmailContactFactory.create())

        if isinstance(self, dict):
            self["email_to"] = [EmailContactFactory.create().id]
            self["email_cc"] = [EmailContactFactory.create().id]
            self["email_bcc"] = [EmailContactFactory.create().id]

    @factory.post_generation
    def exempt_users(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for user in extracted:
                self.exempt_users.add(user)

    default_currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyFactory")


class CompanyAccountingFactory(CompanyFactory):
    entry_accounting_information = factory.RelatedFactory(
        "wbaccounting.factories.EntryAccountingInformationFactory", "entry"
    )



from .booking_entry import BOOKINGENTRY_MENUITEM
from .entry_accounting_information import ENTRYACCOUNTINGINFORMATION_MENUITEM
from .invoice import INVOICE_MENUITEM
from .invoice_type import INVOICETYPE_MENUITEM
from .transactions import TRANSACTION_MENUITEM
from .cashflows import CASHFLOW_MENUITEM



# Generated by Django 5.0.8 on 2024-09-18 08:57

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbaccounting", "0011_transaction"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="external_invoice_users",
            field=models.ManyToManyField(
                blank=True,
                help_text="External users who are able to see the invoices generated for this counterparty",
                related_name="external_accounting_information",
                to=settings.AUTH_USER_MODEL,
                verbose_name="External User",
            ),
        ),
    ]



# Generated by Django 5.0.3 on 2024-04-15 09:05

import django.db.models.deletion
import django_fsm
import wbaccounting.models.entry_accounting_information
from django.conf import settings
from django.db import migrations, models
from django.db.models import Max
from django.utils.html import strip_tags


def swap_net_and_gross(apps, schema_editor):
    # We are using the wrong notion for net and gross, we need to swap them
    BookingEntry = apps.get_model("wbaccounting", "BookingEntry")
    for bookingentry in BookingEntry.objects.all():
        bookingentry.net_value, bookingentry.gross_value = bookingentry.gross_value, bookingentry.net_value
        bookingentry.save()


def migrate_dates(apps, schema_editor):
    # We are migrating from/to date from fields to parameters, as they are getting removed
    # Also we are setting the reference date to to date, as it is the last date of the accounting period
    BookingEntry = apps.get_model("wbaccounting", "BookingEntry")
    for bookingentry in BookingEntry.objects.filter(from_date__isnull=False, to_date__isnull=False):
        bookingentry.parameters = {
            "from_date": bookingentry.from_date.strftime("%d.%m.%Y"),
            "to_date": bookingentry.to_date.strftime("%d.%m.%Y"),
        }
        bookingentry.reference_date = bookingentry.to_date
        bookingentry.save()

    # As the reference date is set for the bookings entries, the
    Invoice = apps.get_model("wbaccounting", "Invoice")
    for invoice in Invoice.objects.filter(booking_entries__isnull=False):
        invoice.reference_date = invoice.booking_entries.aggregate(mv=Max("reference_date")).get("mv")
        invoice.save()


def strip_html_from_subject(apps, schema_editor):
    EntryAccountingInformation = apps.get_model("wbaccounting", "EntryAccountingInformation")

    for eai in EntryAccountingInformation.objects.all():
        eai.email_subject = strip_tags(eai.email_subject)
        eai.save()


def save_models(apps, schema_editor):
    from wbaccounting.models import BookingEntry

    for bookingentry in BookingEntry.objects.all():
        bookingentry.save()


class Migration(migrations.Migration):
    dependencies = [
        ("currency", "0001_initial"),
        ("directory", "0007_alter_bankingcontact_options"),
        ("wbaccounting", "0008_alter_invoice_options"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="InvoiceType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100, unique=True, verbose_name="Name")),
                ("processor", models.CharField(blank=True, max_length=128, null=True, verbose_name="Processor")),
            ],
            options={
                "verbose_name": "Invoice Type",
                "verbose_name_plural": "Invoice Types",
            },
        ),
        migrations.AlterModelOptions(
            name="bookingentry",
            options={"verbose_name": "Booking", "verbose_name_plural": "Bookings"},
        ),
        migrations.AlterModelOptions(
            name="entryaccountinginformation",
            options={"verbose_name": "Accounting Information", "verbose_name_plural": "Accounting Information"},
        ),
        migrations.AlterModelOptions(
            name="invoice",
            options={
                "permissions": (
                    ("can_generate_invoice", "Can Generate Invoice"),
                    ("administrate_invoice", "Can administer Invoice"),
                ),
                "verbose_name": "Invoice",
                "verbose_name_plural": "Invoices",
            },
        ),
        migrations.RemoveField(
            model_name="bookingentry",
            name="accrual_date",
        ),
        migrations.RemoveField(
            model_name="bookingentry",
            name="calculated_value",
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="parameters",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="reference_date",
            field=models.DateField(blank=True, null=True, verbose_name="Reference Date"),
        ),
        migrations.AddField(
            model_name="invoice",
            name="reference_date",
            field=models.DateField(blank=True, null=True, verbose_name="Reference Date"),
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(
            swap_net_and_gross,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.RunPython(
            migrate_dates,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.RemoveField(
            model_name="bookingentry",
            name="from_date",
        ),
        migrations.RemoveField(
            model_name="bookingentry",
            name="to_date",
        ),
        migrations.RemoveField(
            model_name="bookingentry",
            name="main_booking_entry",
        ),
        migrations.RemoveField(
            model_name="bookingentry",
            name="related_data",
        ),
        migrations.RemoveField(
            model_name="bookingentry",
            name="resolved",
        ),
        migrations.RemoveField(
            model_name="entryaccountinginformation",
            name="label_mapping",
        ),
        migrations.RemoveField(
            model_name="entryaccountinginformation",
            name="social_charges",
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="backlinks",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="due_date",
            field=models.DateField(blank=True, null=True, verbose_name="Due Date"),
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="generator",
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="invoice_fx_rate",
            field=models.DecimalField(
                blank=True, decimal_places=6, max_digits=20, null=True, verbose_name="Invoice FX Rate"
            ),
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="invoice_gross_value",
            field=models.DecimalField(
                blank=True, decimal_places=4, max_digits=16, null=True, verbose_name="Invoice Gross Value"
            ),
        ),
        migrations.AddField(
            model_name="bookingentry",
            name="invoice_net_value",
            field=models.DecimalField(
                blank=True, decimal_places=4, max_digits=16, null=True, verbose_name="Invoice Net Value"
            ),
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="booking_entry_generator",
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="counterparty_is_private",
            field=models.BooleanField(
                default=False,
                help_text="Hides all of the counterparty's invoices from non-eligible users",
                verbose_name="Counterparty Is Private",
            ),
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="exempt_users",
            field=models.ManyToManyField(
                blank=True,
                help_text="Exclusion list of users who are able to see private invoices for the counterparty",
                related_name="private_accounting_information",
                to=settings.AUTH_USER_MODEL,
                verbose_name="Exempt Users",
            ),
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="send_mail",
            field=models.BooleanField(default=True, verbose_name="Send Mail"),
        ),
        migrations.AddField(
            model_name="invoice",
            name="backlinks",
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="invoice",
            name="gross_value",
            field=models.DecimalField(
                blank=True, decimal_places=4, max_digits=16, null=True, verbose_name="Gross Value"
            ),
        ),
        migrations.AddField(
            model_name="invoice",
            name="net_value",
            field=models.DecimalField(
                blank=True, decimal_places=4, max_digits=16, null=True, verbose_name="Net Value"
            ),
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="default_currency",
            field=models.ForeignKey(
                blank=True,
                default=wbaccounting.models.entry_accounting_information.default_currency,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="entry_accounting_informations",
                to="currency.currency",
                verbose_name="Default Currency",
            ),
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="email_body",
            field=models.TextField(
                default=wbaccounting.models.entry_accounting_information.default_email_body, verbose_name="Body"
            ),
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="entry",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="entry_accounting_information",
                to="directory.entry",
                verbose_name="Counterparty",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="status",
            field=django_fsm.FSMField(
                choices=[
                    ("DRAFT", "Draft"),
                    ("SUBMITTED", "Submitted"),
                    ("APPROVED", "Approved"),
                    ("CANCELLED", "Cancelled"),
                    ("SENT", "Sent"),
                    ("PAID", "Paid"),
                ],
                default="DRAFT",
                max_length=50,
                verbose_name="Status",
            ),
        ),
        migrations.AddField(
            model_name="entryaccountinginformation",
            name="default_invoice_type",
            field=models.ForeignKey(
                blank=True,
                help_text="When invoicinging outstanding booking entries, this invoice type will be assigned to the corresponding invoice",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="booking_entries",
                to="wbaccounting.invoicetype",
                verbose_name="Default Invoice Type",
            ),
        ),
        migrations.AddField(
            model_name="invoice",
            name="invoice_type",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="invoices",
                to="wbaccounting.invoicetype",
                verbose_name="Type",
            ),
        ),
        migrations.RenameField(
            model_name="bookingentry",
            old_name="resolved_date",
            new_name="payment_date",
        ),
        migrations.AlterField(
            model_name="bookingentry",
            name="payment_date",
            field=models.DateField(blank=True, null=True, verbose_name="Payment Date"),
        ),
        migrations.RemoveField(
            model_name="invoice",
            name="is_counterparty_invoice",
        ),
        migrations.AddField(
            model_name="invoice",
            name="payment_date",
            field=models.DateField(blank=True, null=True, verbose_name="Payment Date"),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="invoice_date",
            field=models.DateField(verbose_name="Invoice Date"),
        ),
        migrations.AlterModelOptions(
            name="entryaccountinginformation",
            options={"verbose_name": "Counterparty", "verbose_name_plural": "Counterparties"},
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="email_subject",
            field=models.CharField(default="{{invoice.title}}", max_length=1024, verbose_name="Subject"),
        ),
        migrations.AlterField(
            model_name="entryaccountinginformation",
            name="entry",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="entry_accounting_information",
                to="directory.entry",
                verbose_name="Linked Counterparty",
            ),
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(
            strip_html_from_subject,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(
            save_models,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
    ]



import pytest


@pytest.mark.django_db
def test_dummy_processor(invoice, mocker):
    from wbaccounting.processors import dummy_processor

    invoice_type = invoice.invoice_type
    invoice_type.processor = "wbaccounting.processors.dummy_processor.processor"
    invoice_type.save()
    invoice.status = invoice.Status.SUBMITTED
    invoice.save()

    dummy_processor = mocker.spy(dummy_processor, "processor")
    invoice.approve()
    invoice.save()
    dummy_processor.assert_called_once()
    assert invoice.status == invoice.Status.SENT


@pytest.mark.django_db
def test_no_processor(invoice):
    invoice.status = invoice.Status.SUBMITTED
    invoice.save()

    invoice.approve()
    invoice.save()
    assert invoice.status == invoice.Status.APPROVED



from django.contrib import admin

from wbaccounting.models import Transaction


@admin.register(Transaction)
class TransactionModelAdmin(admin.ModelAdmin):
    list_display = ["booking_date", "value_date", "value", "bank_account", "prenotification"]

    raw_id_fields = ["import_source"]

    autocomplete_fields = [
        "bank_account",
        "from_bank_account",
        "to_bank_account",
        "currency",
    ]



from contextlib import suppress
from decimal import Decimal
from io import BytesIO
from typing import TYPE_CHECKING

from django.contrib.auth import get_user_model
from django.core.files import File
from django.db import models, transaction
from django.db.models import Max, Q, QuerySet, Sum
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.template import Context, Template
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from dynamic_preferences.registries import global_preferences_registry
from slugify import slugify
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models import Company, Entry
from wbcore.contrib.documents.models import Document, DocumentType
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.models import WBModel
from wbcore.utils.importlib import import_from_dotted_path

from wbaccounting.dynamic_preferences_registry import format_invoice_number
from wbaccounting.files.invoice_document_file import generate_file
from wbaccounting.models.booking_entry import BookingEntry, BookingEntryDefaultQuerySet
from wbaccounting.models.model_tasks import (
    refresh_complete_invoice_as_task,
    refresh_invoice_document_as_task,
)


class InvoiceDefaultQuerySet(QuerySet):
    def create_for_booking_entry(self, booking_entry: BookingEntry, **kwargs) -> "Invoice":
        """
        Creates an Invoice instance associated with a specific BookingEntry and updates the BookingEntry
        to link to this newly created Invoice.

        This method takes a BookingEntry object as input, creates a new Invoice associated with the BookingEntry's
        counterparty, and then links the BookingEntry to this new Invoice by updating its 'invoice' field.

        Args:
            booking_entry (BookingEntry): The BookingEntry instance for which to create the invoice.

        Returns:
            Invoice: The newly created Invoice instance associated with the given BookingEntry's counterparty.
        """
        invoice = super().create(
            counterparty=booking_entry.counterparty,
            invoice_type=booking_entry.counterparty.entry_accounting_information.default_invoice_type,
            **kwargs,
        )
        booking_entry.invoice = invoice
        booking_entry.save()
        return invoice

    def create_for_counterparty(self, counterparty: Entry, **kwargs) -> "Invoice|None":
        """
        Creates an Invoice instance for a given counterparty and associates unresolved BookingEntry objects with the
        newly created invoice.

        This method automatically finds all BookingEntry objects related to the counterparty that do not have an
        associated invoice and are not resolved, and updates them to link to the newly created Invoice.

        Args:
            counterparty (Entry): The counterparty for which to create the invoice.

        Returns:
            Invoice | None: The newly created Invoice instance associated with the given counterparty or None if no
                            Booking Entries can be invoiced.
        """
        booking_entries = BookingEntry.objects.filter(
            counterparty=counterparty,
            invoice__isnull=True,
            payment_date__isnull=True,
        )
        if not booking_entries.exists():
            return None

        title = kwargs.pop(
            "title",
            f"Invoice {counterparty.computed_str} {booking_entries.latest('reference_date').reference_date or ''}",
        )
        invoice_currency = kwargs.pop("invoice_currency", counterparty.entry_accounting_information.default_currency)
        invoice_type = kwargs.pop("type", counterparty.entry_accounting_information.default_invoice_type)
        backlinks = counterparty.entry_accounting_information.get_booking_entry_generator().merge_backlinks(
            booking_entries
        )

        invoice = super().create(
            counterparty=counterparty,
            title=title,
            invoice_currency=invoice_currency,
            backlinks=backlinks,
            invoice_type=invoice_type,
            **kwargs,
        )
        booking_entries.update(invoice=invoice)
        transaction.on_commit(lambda: refresh_complete_invoice_as_task.delay(invoice.pk))  # type: ignore
        return invoice

    def filter_for_user(self, user: User) -> QuerySet:
        """
        Filters invoices based on if the current user can see the counterparty.

        Args:
            user (User): The user for whom invoices need to be filtered.

        Returns:
            QuerySet: A filtered queryset.
        """

        if user.is_superuser or user.has_perm("wbaccounting.administrate_invoice"):
            return self

        if not user.has_perm("wbaccounting.view_invoice"):
            return self.none()

        return self.filter(
            Q(counterparty__entry_accounting_information__counterparty_is_private=False)
            | Q(counterparty__entry_accounting_information__exempt_users=user)
        )


class Invoice(WBModel):
    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        SUBMITTED = "SUBMITTED", "Submitted"
        APPROVED = "APPROVED", "Approved"
        CANCELLED = "CANCELLED", "Cancelled"
        SENT = "SENT", "Sent"
        PAID = "PAID", "Paid"

    class Meta:
        verbose_name = "Invoice"
        verbose_name_plural = "Invoices"
        permissions = (
            (
                "can_generate_invoice",
                "Can Generate Invoice",
            ),
            ("administrate_invoice", "Can administer Invoice"),
        )

        notification_types = [
            create_notification_type(
                "wbaccounting.invoice.notify_approval",
                "Invoice Approval Notification",
                "Sends a notification when something happens in a relevant article.",
                True,
                True,
                False,
            ),
        ]

    def __str__(self):
        return self.title

    objects = InvoiceDefaultQuerySet.as_manager()

    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name="Status")
    resolved = models.BooleanField(default=False, verbose_name="Resolved")

    title = models.CharField(max_length=255, verbose_name="Title")

    invoice_date = models.DateField(verbose_name="Invoice Date")
    reference_date = models.DateField(verbose_name="Reference Date", null=True, blank=True)
    payment_date = models.DateField(verbose_name="Payment Date", null=True, blank=True)

    invoice_currency = models.ForeignKey(
        "currency.Currency", related_name="invoices", on_delete=models.PROTECT, verbose_name="Curency"
    )

    counterparty = models.ForeignKey(
        "directory.Entry", related_name="accounting_invoices", on_delete=models.PROTECT, verbose_name="Counterparty"
    )

    text_above = models.TextField(null=True, blank=True)
    text_below = models.TextField(null=True, blank=True)

    invoice_type = models.ForeignKey(
        to="InvoiceType", null=True, verbose_name="Type", related_name="invoices", on_delete=models.PROTECT
    )
    gross_value = models.DecimalField(
        max_digits=16, decimal_places=4, null=True, blank=True, verbose_name="Gross Value"
    )
    net_value = models.DecimalField(max_digits=16, decimal_places=4, null=True, blank=True, verbose_name="Net Value")

    backlinks = models.JSONField(null=True, blank=True)

    if TYPE_CHECKING:
        booking_entries: BookingEntryDefaultQuerySet

    def __init__(self, *args, **kwargs):
        super(Invoice, self).__init__(*args, **kwargs)
        if hasattr(self, "invoice_currency"):  # FDM instantiate model without any attribute
            self._original_invoice_currency = self.invoice_currency

    def save(self, *args, **kwargs):
        # If the invoice does not have a primary key yet, then it can't be set as a booking entries invoice
        if self.id:
            self.gross_value = self.booking_entries.all().aggregate(s=Sum("invoice_gross_value")).get("s", Decimal(0))
            self.net_value = self.booking_entries.all().aggregate(s=Sum("invoice_net_value")).get("s", Decimal(0))
            self.reference_date = self.booking_entries.all().aggregate(m=Max("reference_date")).get("m", None)

            transaction.on_commit(lambda: refresh_invoice_document_as_task.delay(self.pk))  # type: ignore
        super().save(*args, **kwargs)
        if self.invoice_currency != self._original_invoice_currency:
            refresh_complete_invoice_as_task.delay(self.id)

    @property
    def is_counterparty_invoice(self):
        return (self.net_value or 0) < 0

    @property
    def invoice_system_key(self) -> str:
        return f"invoice-{self.pk}"

    @property
    def invoice_document(self):
        return Document.get_for_object(self).filter(system_created=True, system_key=self.invoice_system_key).first()

    @property
    def invoice_company(self):
        global_preferences = global_preferences_registry.manager()
        if invoice_company_id := global_preferences["wbaccounting__invoice_company"]:
            return Company.objects.get(id=invoice_company_id)

    @transition(
        field=status,
        source=[Status.DRAFT],  # type: ignore
        target=Status.SUBMITTED,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbaccounting:invoice",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label="Submit",
                action_label="Submitting",
                description_fields="<p>{{title}}</p><p>After Submitting, this invoice cannot be changed anymore.</p>",
            )
        },
    )
    def submit(self, by=None, description=None, **kwargs):
        for user in get_user_model().objects.filter(
            Q(user_permissions__codename="administrate_invoice")
            | Q(groups__permissions__codename="administrate_invoice")
        ):
            send_notification(
                code="wbaccounting.invoice.notify_approval",
                title="An invoice needs to be approved",
                body=f"An Invoice was submitted for approval ({self.title})",
                user=user,
                reverse_name="wbaccounting:invoice-detail",
                reverse_args=[self.pk],
            )

    @transition(
        field=status,
        source=[Status.SUBMITTED, Status.DRAFT],  # type: ignore
        target=Status.CANCELLED,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbaccounting:invoice",),
                icon=WBIcon.REJECT.icon,
                key="cancel",
                label="Cancel",
                action_label="Cancellation",
                description_fields="<p>{{title}}</p><p>After cancelling, this invoice cannot be used anymore.</p>",
            )
        },
    )
    def cancel(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.SUBMITTED],  # type: ignore
        target=Status.APPROVED,
        permission=lambda instance, user: user.has_perm("wbaccounting.administrate_invoice"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbaccounting:invoice",),
                icon=WBIcon.APPROVE.icon,
                key="approve",
                label="Approve",
                action_label="Approval",
                description_fields="<p>Are you sure you want to approve this invoice?</p>",
            )
        },
    )
    def approve(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.SUBMITTED],  # type: ignore
        target=Status.DRAFT,
        permission=lambda instance, user: user.has_perm("wbaccounting.administrate_invoice"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbaccounting:invoice",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Denial",
                description_fields="<p>Are you sure you want to deny this invoice?</p>",
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.APPROVED],  # type: ignore
        target=Status.SENT,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.SUCCESS,
                identifiers=("wbaccounting:invoice",),
                icon=WBIcon.GENERATE_NEXT.icon,
                key="send",
                label="Send",
                action_label="Sending",
                description_fields="<p>{{title}}</p>",
            )
        },
    )
    def send(self, by=None, description=None, **kwargs):
        pass

    def can_send(self):
        errors = dict()
        if not self.invoice_document:
            errors["status"] = [_("An invoice needs to be generated first.")]
        return errors

    @transition(
        field=status,
        source=[Status.SENT],  # type: ignore
        target=Status.PAID,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.SUCCESS,
                identifiers=("wbaccounting:invoice",),
                icon=WBIcon.DEAL_MONEY.icon,
                key="pay",
                label="Pay",
                action_label="Payment",
                description_fields="<p>{{title}}</p>",
            )
        },
    )
    def pay(self, by=None, description=None, **kwargs):
        pass

    def send_invoice_to_recipients(self):
        if invoice_document := self.invoice_document:
            entry_accounting_information = self.counterparty.entry_accounting_information
            if entry_accounting_information.send_mail:
                global_preferences = global_preferences_registry.manager()
                from_email = global_preferences["wbaccounting__default_from_email_address"]
                context = {"invoice": self.get_context(), "entry": self.counterparty}

                rendered_subject = Template(entry_accounting_information.email_subject).render(Context(context))
                rendered_body = Template(entry_accounting_information.email_body).render(Context(context))

                to_emails = list(entry_accounting_information.email_to.values_list("address", flat=True))
                cc_emails = list(entry_accounting_information.email_cc.values_list("address", flat=True))
                bcc_emails = list(entry_accounting_information.email_bcc.values_list("address", flat=True))

                invoice_document.send_email(
                    to_emails,
                    as_link=False,
                    subject=rendered_subject,
                    from_email=from_email,
                    body=rendered_body,
                    cc_emails=cc_emails,
                    bcc_emails=bcc_emails,
                )

    # Other methods
    def refresh_invoice_document(self, override_status=False):
        if self.status == self.Status.DRAFT or override_status:
            invoice = generate_file(self)
            file_name = f"{slugify(self.title)}.pdf"
            document_file = BytesIO(invoice)
            document_type, created = DocumentType.objects.get_or_create(name="invoice")
            document, created = Document.objects.update_or_create(
                document_type=document_type,
                system_created=True,
                system_key=self.invoice_system_key,
                defaults={
                    "name": f"Invoice: {self.title}",
                    "description": f"Invoice for {self.counterparty}.",
                    "file": File(document_file, file_name),
                },
            )
            document.link(self)

    def get_permissions_for_user_and_document(self, user, document, created=None) -> list[tuple[str, bool]]:
        # if the document is the invoice document, we want to assign certain permissions
        if document.system_created and self.invoice_system_key == document.system_key:
            if user.is_superuser:
                return []

            entry_info = self.counterparty.entry_accounting_information
            has_view_permission = user.has_perm("wbaccounting.view_invoice")

            if entry_info.external_invoice_users.filter(id=user.id).exists():
                return [("documents.view_document", False)]

            is_access_allowed = has_view_permission and (
                not entry_info.counterparty_is_private or user in entry_info.exempt_users.all()
            )
            has_admin_permission = user.has_perm("wbaccounting.administrate_invoice")

            if is_access_allowed or has_admin_permission:
                return [("documents.view_document", False)]
        return []

    def get_context(self) -> dict:
        return {
            "title": self.title,
            "counterparty": self.counterparty.computed_str,
            "total_net_value": format_invoice_number(self.net_value),
            "total_gross_value": format_invoice_number(self.gross_value),
            "text_above": self.text_above,
            "text_below": self.text_below,
            "invoice_date": self.invoice_date.strftime("%d.%m.%Y"),
            "currency": self.invoice_currency,
            "invoice": self,
        }

    @classmethod
    def get_endpoint_basename(cls):
        return "wbaccounting:invoice"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbaccounting:invoicerepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}} ({{invoice_date}})"


@receiver(pre_save, sender=Invoice)
def post_save_handle_processor(sender, instance, **kwargs):
    if (
        instance.status == Invoice.Status.APPROVED
        and instance.invoice_type
        and (processor_path := instance.invoice_type.processor)
    ):
        with suppress(ModuleNotFoundError):
            processor = import_from_dotted_path(processor_path)
            processor(instance)
            instance.status = instance.Status.SENT



from decimal import Decimal

import pytest
from pytest_factoryboy import LazyFixture
from wbcore.contrib.authentication.models import User

from wbaccounting.models import BookingEntry, EntryAccountingInformation, Invoice


@pytest.mark.django_db
class TestBookingEntry:
    def test_str(self, booking_entry: BookingEntry):
        assert str(booking_entry) == booking_entry.title

    @pytest.mark.parametrize(
        "method,return_value",
        [
            ("get_endpoint_basename", "wbaccounting:bookingentry"),
            ("get_representation_value_key", "id"),
            ("get_representation_label_key", "{{title}}"),
            ("get_representation_endpoint", "wbaccounting:bookingentryrepresentation-list"),
        ],
    )
    def test_wbmodel_methods(self, method: str, return_value: str):
        assert getattr(BookingEntry, method)() == return_value

    @pytest.mark.parametrize("user__is_superuser", [False])
    def test_filter_for_user_no_superuser(self, booking_entry: BookingEntry, user: User):
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry not in booking_entries

    @pytest.mark.parametrize("user__is_superuser", [True])
    def test_filter_for_user_superuser(self, booking_entry: BookingEntry, user: User):
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry in booking_entries

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_bookingentry"])])
    @pytest.mark.parametrize("booking_entry__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [False])
    def test_filter_for_user_public_counterparty(
        self, booking_entry: BookingEntry, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry in booking_entries

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_bookingentry"])])
    @pytest.mark.parametrize("booking_entry__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty(
        self, booking_entry: BookingEntry, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry not in booking_entries

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_bookingentry"])])
    @pytest.mark.parametrize("booking_entry__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty_with_exempt(
        self, booking_entry: BookingEntry, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        entry_accounting_information.exempt_users.add(user)
        booking_entries = BookingEntry.objects.filter_for_user(user)  # type: ignore
        assert booking_entry in booking_entries

    @pytest.mark.parametrize("booking_entry__net_value", [Decimal(90)])
    @pytest.mark.parametrize("booking_entry__gross_value", [None])
    @pytest.mark.parametrize("booking_entry__vat", [Decimal(0.1)])
    def test_net_without_gross(self, booking_entry: BookingEntry):
        assert pytest.approx(booking_entry.gross_value) == Decimal(99)

    @pytest.mark.parametrize("booking_entry__net_value", [None])
    @pytest.mark.parametrize("booking_entry__gross_value", [Decimal(99)])
    @pytest.mark.parametrize("booking_entry__vat", [Decimal(0.1)])
    def test_gross_without_net(self, booking_entry: BookingEntry):
        assert pytest.approx(booking_entry.net_value) == Decimal(90)

    @pytest.mark.parametrize("booking_entry__net_value", [Decimal(90)])
    @pytest.mark.parametrize("booking_entry__gross_value", [Decimal(100)])
    @pytest.mark.parametrize("booking_entry__vat", [Decimal(0.1)])
    def test_net_with_gross(self, booking_entry: BookingEntry):
        assert pytest.approx(booking_entry.gross_value) == Decimal(99)

    @pytest.mark.parametrize("booking_entry__net_value", [None])
    @pytest.mark.parametrize("booking_entry__gross_value", [None])
    @pytest.mark.parametrize("booking_entry__vat", [Decimal(0.1)])
    def test_net_and_gross_none(self, booking_entry: BookingEntry):
        assert booking_entry.net_value == Decimal(0) and booking_entry.gross_value == Decimal(0)

    def test_invoice_value_same_currency(self, booking_entry: BookingEntry):
        assert booking_entry.net_value == booking_entry.invoice_net_value
        assert booking_entry.gross_value == booking_entry.invoice_gross_value

    def test_invoice_value_different_currency(
        self, booking_entry: BookingEntry, invoice: Invoice, currency_factory, mocker
    ):
        # Change the invoice currency
        invoice.invoice_currency = currency_factory.create()
        invoice.save()

        # Patch the currency convert method to no rely on the database
        mocker.patch("wbcore.contrib.currency.models.Currency.convert", return_value=Decimal(1.1))
        # Run the save method to trigger fx computation
        booking_entry.save()
        assert booking_entry.invoice_net_value == pytest.approx(
            booking_entry.net_value * booking_entry.invoice_fx_rate
        )  # type: ignore

    def test_invoice_save_on_change(self, booking_entry: BookingEntry, mocker):
        mocked_save = mocker.patch("wbaccounting.models.invoice.Invoice.save", autospec=True)
        booking_entry.save()
        mocked_save.assert_called_once()

    def test_invoice_delete_on_change(self, booking_entry: BookingEntry, mocker):
        mocked_save = mocker.patch("wbaccounting.models.invoice.Invoice.save", autospec=True)
        booking_entry.delete()
        mocked_save.assert_called_once()



# Generated by Django 5.0.3 on 2024-04-29 11:33

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbaccounting", "0009_invoicetype_alter_bookingentry_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="bookingentry",
            options={
                "permissions": (("can_generate_booking_entries", "Can Generate Bookings"),),
                "verbose_name": "Booking",
                "verbose_name_plural": "Bookings",
            },
        ),
    ]



# Generated by Django 4.2.9 on 2024-02-20 10:08

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbaccounting", "0007_alter_invoice_options"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="invoice",
            options={
                "permissions": (("administrate_invoice", "Can Administrate Invoice"),),
                "verbose_name": "Invoice",
                "verbose_name_plural": "Invoices",
            },
        ),
    ]



from wbcore import viewsets

from wbaccounting.models import InvoiceType
from wbaccounting.serializers import (
    InvoiceTypeModelSerializer,
    InvoiceTypeRepresentationSerializer,
)
from wbaccounting.viewsets.display import InvoiceTypeDisplayConfig
from wbaccounting.viewsets.titles import InvoiceTypeTitleConfig


class InvoiceTypeModelViewSet(viewsets.ModelViewSet):
    queryset = InvoiceType.objects.all()
    serializer_class = InvoiceTypeModelSerializer
    search_fields = ordering_fields = ("name", "processor")
    ordering = ("name", "id")
    display_config_class = InvoiceTypeDisplayConfig
    filterset_fields = {"name": ["exact", "icontains"], "processor": ["exact", "icontains"]}
    title_config_class = InvoiceTypeTitleConfig


class InvoiceTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = InvoiceTypeRepresentationSerializer
    queryset = InvoiceType.objects.all()
    search_fields = ordering_fields = ("name",)
    ordering = ("name", "id")



from decimal import Decimal
from typing import TYPE_CHECKING

from django.db import models
from django.utils.translation import gettext as _
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.models import WBModel

from wbaccounting.io.handlers.transactions import TransactionImportHandler

if TYPE_CHECKING:
    from wbcore.contrib.authentication.models import User


class TransactionQuerySet(models.QuerySet):
    def filter_for_user(self, user: "User") -> models.QuerySet["Transaction"]:
        if user.is_superuser:
            return self

        return self.filter(bank_account__access__user_account__in=[user])


class Transaction(ImportMixin, WBModel):
    """A transaction represents a bank transfer of some funds in some currency from one Bank Account to another one."""

    import_export_handler_class = TransactionImportHandler

    booking_date = models.DateField(verbose_name=_("Booking Date"))
    value_date = models.DateField(verbose_name=_("Value Date"))

    bank_account = models.ForeignKey(
        to="directory.BankingContact",
        related_name="wbaccounting_transactions",
        on_delete=models.PROTECT,
        verbose_name=_("Linked Bank Account"),
    )
    from_bank_account = models.ForeignKey(
        to="directory.BankingContact",
        related_name="+",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name=_("Source Bank Account"),
    )
    to_bank_account = models.ForeignKey(
        to="directory.BankingContact",
        related_name="+",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name=_("Target Bank Account"),
    )
    currency = models.ForeignKey(
        to="currency.Currency",
        related_name="+",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name=_("Currency"),
    )
    fx_rate = models.DecimalField(default=Decimal(1), max_digits=10, decimal_places=4, verbose_name=_("FX Rate"))
    value_local_ccy = models.DecimalField(
        max_digits=19, decimal_places=2, null=True, blank=True, verbose_name=_("Value (Local Currency)")
    )
    value = models.DecimalField(max_digits=19, decimal_places=2, null=True, blank=True, verbose_name=_("Value"))
    description = models.TextField(default="")
    prenotification = models.BooleanField(
        default=False,
        verbose_name=_("Prenotification"),
        help_text=_("This field indicates that this transaction will happen sometime in the future."),
    )
    _hash = models.CharField(max_length=64, null=True, blank=True)

    objects = TransactionQuerySet.as_manager()

    def save(self, *args, **kwargs):
        # We make sure that the relationship between value and
        # value_local_ccy stays consistant while prefering value
        if self.value is not None:
            self.value_local_ccy = self.value / self.fx_rate

        elif self.value_local_ccy is not None:
            self.value = self.value_local_ccy * self.fx_rate

        # If value date is not set we set it to the booking date
        if self.value_date is None:
            self.value_date = self.booking_date

        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return f"{self.booking_date:%d.%m.%Y}: {self.value:.2f}"

    class Meta:
        default_related_name = "transactions"
        verbose_name = _("Transaction")
        verbose_name_plural = _("Transactions")

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbaccounting:transaction"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbaccounting:transactionrepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{booking_date}}: {{value}}"



from rest_framework.reverse import reverse
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class InvoiceBaseButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        return {bt.HyperlinkButton(key="invoice_file", label="Invoice", icon=WBIcon.DOCUMENT.icon)}

    def get_custom_list_instance_buttons(self):
        return self.get_custom_instance_buttons()


class InvoiceButtonConfig(InvoiceBaseButtonConfig):
    def get_custom_buttons(self) -> set:
        buttons = super().get_custom_buttons()
        if not (self.view.kwargs.get("pk") or self.new_mode):
            buttons.add(
                bt.WidgetButton(
                    label="Consolidated Invoices",
                    endpoint=reverse("wbaccounting:consolidated-invoice-list", args=[], request=self.request),
                )
            )
        return buttons


class ConsolidatedInvoiceButtonConfig(InvoiceBaseButtonConfig):
    def get_custom_list_instance_buttons(self) -> set:
        return {
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbaccounting:invoice",),
                label="Approve ({{num_submitted}})",
                description_fields="""
                        <p>Do you want to approve all {{num_submitted}} submitted invoices?</p>
                        """,
                action_label="Approving",
                title="Approve",
                key="approve",
            ),
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbaccounting:invoice",),
                label="Submit ({{num_draft}})",
                description_fields="""
                        <p>Do you want to submit {{num_draft}} invoices?</p>
                        """,
                action_label="Submitting",
                title="Submit",
                key="submit",
            ),
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbaccounting:invoice",),
                label="Pay ({{num_sent}})",
                description_fields="""
                        <p>Do you want to pay all {{num_sent}} invoices?</p>
                        """,
                action_label="Payment",
                title="Pay",
                key="pay",
            ),
        }



import factory
import pytest

from wbaccounting.models import EntryAccountingInformation
from wbaccounting.serializers import (
    EntryAccountingInformationModelSerializer,
    EntryAccountingInformationRepresentationSerializer,
)


@pytest.mark.django_db
class TestEntryAccountingInformationModelSerializer:
    def test_serialize(self, entry_accounting_information: EntryAccountingInformation):
        serializer = EntryAccountingInformationModelSerializer(entry_accounting_information)
        assert isinstance(serializer.data, dict)

    #
    def test_serialize_contains_keys(self, entry_accounting_information: EntryAccountingInformation):
        serializer = EntryAccountingInformationModelSerializer(entry_accounting_information)
        assert (
            "id",
            "entry",
            "_entry",
            "tax_id",
            "vat",
            "default_currency",
            "_default_currency",
            "default_invoice_type",
            "_default_invoice_type",
            "email_to",
            "email_cc",
            "email_bcc",
            "_email_to",
            "_email_cc",
            "_email_bcc",
            "email_subject",
            "email_body",
            "send_mail",
            "counterparty_is_private",
            "exempt_users",
            "_exempt_users",
            "booking_entry_generator",
            "external_invoice_users",
            "_external_invoice_users",
            "_additional_resources",
        ) == tuple(serializer.data.keys())  # type: ignore

    def test_deserialize(self, entry_accounting_information_factory, entry, currency):
        data = factory.build(dict, FACTORY_CLASS=entry_accounting_information_factory)
        data["entry"] = entry.pk
        data["default_currency"] = currency.pk
        serializer = EntryAccountingInformationModelSerializer(data=data)
        assert serializer.is_valid()


@pytest.mark.django_db
class TestEntryAccountingInformationRepresentationSerializer:
    def test_serialize(self, entry_accounting_information: EntryAccountingInformation):
        serializer = EntryAccountingInformationRepresentationSerializer(entry_accounting_information)
        assert serializer.data == {
            "id": entry_accounting_information.pk,
            "entry_repr": entry_accounting_information.entry.computed_str,
        }



import factory
from factory.fuzzy import FuzzyDecimal

from wbaccounting.models import BookingEntry


class BookingEntryFactory(factory.django.DjangoModelFactory):
    class Meta:  # type: ignore
        model = BookingEntry

    # resolved
    title = factory.Faker("text", max_nb_chars=64)
    booking_date = factory.Faker("date_between", start_date="+2d", end_date="+3d")
    payment_date = factory.Faker("date_object")
    reference_date = factory.Faker("date_object")
    gross_value = factory.Faker("pydecimal", right_digits=4, min_value=0, max_value=99999999999)
    net_value = factory.Faker("pydecimal", right_digits=4, min_value=0, max_value=99999999999)
    vat = FuzzyDecimal(0, 0.9, 4)
    currency = factory.SelfAttribute("invoice.invoice_currency")

    invoice = factory.SubFactory("wbaccounting.factories.InvoiceFactory")
    counterparty = factory.SubFactory("wbcore.contrib.directory.factories.EntryFactory")



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

BOOKINGENTRY_MENUITEM = MenuItem(
    label="Bookings",
    endpoint="wbaccounting:bookingentry-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbaccounting.view_bookingentry"]
    ),
    add=MenuItem(
        label="Create Booking",
        endpoint="wbaccounting:bookingentry-list",
        permission=ItemPermission(permissions=["wbaccounting.add_bookingentry"]),
    ),
)



from django.contrib import admin

from wbaccounting.models import InvoiceType


@admin.register(InvoiceType)
class InvoiceTypeModelAdmin(admin.ModelAdmin):
    search_fields = ("name", "processor")
    list_display = ("id", "name", "processor")



import csv
import datetime
import hashlib
from io import TextIOWrapper

from schwifty import IBAN

# Bank specific Information for Societe Generale Luxembourg
BANK_COUNTRY_CODE = "LU"
BANK_CODE = "060"


def parse(import_source):
    # Load file into a CSV DictReader
    csv_file = import_source.file.open(mode="rb")
    wrapped_file = TextIOWrapper(csv_file, encoding="utf-8", errors="ignore")
    # Read file into a CSV Dict Reader
    csv_reader = csv.DictReader(wrapped_file, delimiter=";")

    # Iterate through the CSV File and parse the data into a list
    data = list()
    for transaction in csv_reader:
        if amount := transaction.get("Amount", None):
            booking_date = datetime.datetime.strptime(transaction["Accounting date"], "%Y/%m/%d").date()
            value_date = datetime.datetime.strptime(transaction["Value date"], "%Y/%m/%d").date()
            currency = transaction["Account currency"]
            value = float(amount.replace(",", "."))
            description = transaction["Transaction main description"]
            bank_account = str(
                IBAN.generate(BANK_COUNTRY_CODE, bank_code=BANK_CODE, account_code=transaction["Account number"])
            )
            item = {
                "booking_date": booking_date.strftime("%Y-%m-%d"),
                "value_date": value_date.strftime("%Y-%m-%d"),
                "currency": currency,
                "value": value,
                "description": description,
                "bank_account": bank_account,
            }
            _hash = hashlib.sha256()
            for field in item.values():
                _hash.update(str(field).encode())

            item["_hash"] = _hash.hexdigest()
            data.append(item)

    csv_file.close()
    return {
        "data": data,
    }



from decimal import Decimal

from django.db.models import Q
from django.dispatch import receiver
from rest_framework.exceptions import ValidationError
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.authentication.models import User
from wbcore.contrib.authentication.serializers import UserRepresentationSerializer
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.serializers import (
    CompanyModelSerializer,
    EmailContactRepresentationSerializer,
    EntryModelSerializer,
    EntryRepresentationSerializer,
    PersonModelSerializer,
)
from wbcore.signals import add_instance_additional_resource

from wbaccounting.generators.base import get_all_booking_entry_choices
from wbaccounting.models import EntryAccountingInformation
from wbaccounting.serializers import InvoiceTypeRepresentationSerializer


class EntryAccountingInformationRepresentationSerializer(serializers.RepresentationSerializer):
    entry_repr = serializers.CharField(source="entry.computed_str", read_only=True)

    class Meta:
        model = EntryAccountingInformation
        fields = (
            "id",
            "entry_repr",
        )


class EntryAccountingInformationModelSerializer(serializers.ModelSerializer):
    _entry = EntryRepresentationSerializer(source="entry")
    _default_currency = CurrencyRepresentationSerializer(source="default_currency")
    _default_invoice_type = InvoiceTypeRepresentationSerializer(source="default_invoice_type")
    vat = serializers.DecimalField(percent=True, required=False, max_digits=6, decimal_places=4, default=Decimal(0))
    _exempt_users = UserRepresentationSerializer(source="exempt_users", many=True)

    _email_to = EmailContactRepresentationSerializer(source="email_to", many=True, ignore_filter=True)
    _email_cc = EmailContactRepresentationSerializer(source="email_cc", many=True, ignore_filter=True)
    _email_bcc = EmailContactRepresentationSerializer(source="email_bcc", many=True, ignore_filter=True)
    booking_entry_generator = serializers.ChoiceField(
        choices=list(get_all_booking_entry_choices()), required=False, allow_null=True
    )

    _external_invoice_users = UserRepresentationSerializer(source="external_invoice_users", many=True)

    @serializers.register_only_instance_resource()
    def generate(self, instance, request, user, **kwargs):
        generators = {}

        if instance.booking_entry_generator and (
            user.is_superuser or user.has_perm("wbaccounting.can_generate_booking_entries")
        ):
            generators["generate_booking_entries"] = reverse(
                "wbaccounting:entryaccountinginformation-generate-booking-entries",
                args=[instance.id],
                request=request,
            )

        if user.is_superuser or user.has_perm("wbaccounting.can_generate_invoice"):
            if instance.entry.booking_entries.filter(Q(invoice__isnull=True) & Q(payment_date__isnull=True)).exists():
                generators["invoice_booking_entries"] = reverse(
                    "wbaccounting:entryaccountinginformation-invoice-booking-entries",
                    args=[instance.id],
                    request=request,
                )

        return generators

    @serializers.register_only_instance_resource()
    def inline_lists(self, instance, request, user, **kwargs):
        return {
            "invoices": reverse(
                "wbaccounting:entryaccountinginformation-invoice-list",
                args=[instance.id],
                request=request,
            ),
            "bookingentries": reverse(
                "wbaccounting:entryaccountinginformation-bookingentry-list",
                args=[instance.id],
                request=request,
            ),
        }

    class Meta:
        model = EntryAccountingInformation

        percent_fields = ["vat"]
        fields = (
            "id",
            "entry",
            "_entry",
            "tax_id",
            "vat",
            "default_currency",
            "_default_currency",
            "default_invoice_type",
            "_default_invoice_type",
            "email_to",
            "email_cc",
            "email_bcc",
            "_email_to",
            "_email_cc",
            "_email_bcc",
            "email_subject",
            "email_body",
            "send_mail",
            "counterparty_is_private",
            "exempt_users",
            "_exempt_users",
            "booking_entry_generator",
            "external_invoice_users",
            "_external_invoice_users",
            "_additional_resources",
        )

    def validate(self, data: dict) -> dict:
        counterparty_is_private: bool | None = data.get(
            "counterparty_is_private", self.instance.counterparty_is_private if self.instance else None
        )
        exempt_users: list[User] | None = data.get(
            "exempt_users", list(self.instance.exempt_users.all()) if self.instance else None
        )

        if exempt_users and not counterparty_is_private:
            raise ValidationError({"exempt_users": "You can only select exempt users for private counterparties."})

        return super().validate(data)


@receiver(add_instance_additional_resource, sender=CompanyModelSerializer)
@receiver(add_instance_additional_resource, sender=PersonModelSerializer)
@receiver(add_instance_additional_resource, sender=EntryModelSerializer)
def entry_adding_additional_resource(sender, serializer, instance, request, user, **kwargs):
    if hasattr(instance, "entry_accounting_information"):
        entry_accounting_information = instance.entry_accounting_information
        if entry_accounting_information:
            return {
                "accounting-information": reverse(
                    "wbaccounting:entryaccountinginformation-detail",
                    args=[entry_accounting_information.id],
                    request=request,
                )
            }
    return {}



from wbcore import serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.serializers import BankingContactRepresentationSerializer

from wbaccounting.models import Transaction


class TransactionRepresentationSerializer(serializers.RepresentationSerializer):
    _bank_account = BankingContactRepresentationSerializer(source="bank_account")
    _detail = serializers.HyperlinkField(reverse_name="wbaccounting:transaction-detail")

    class Meta:
        model = Transaction
        fields = ("id", "booking_date", "bank_account", "_bank_account", "value", "_detail")


class TransactionModelSerializer(serializers.ModelSerializer):
    _bank_account = BankingContactRepresentationSerializer(source="bank_account")
    _from_bank_account = BankingContactRepresentationSerializer(source="from_bank_account")
    _to_bank_account = BankingContactRepresentationSerializer(source="to_bank_account")
    _currency = CurrencyRepresentationSerializer(source="currency")
    bank_account_currency_symbol = serializers.CharField(read_only=True)

    class Meta:
        model = Transaction
        decorators = {
            "value": serializers.decorator(
                decorator_type="text", position="left", value="{{bank_account_currency_symbol}}"
            ),
            "value_local_ccy": serializers.decorator(
                decorator_type="text", position="left", value="{{_currency.symbol}}"
            ),
        }
        fields = (
            "id",
            "booking_date",
            "value_date",
            "bank_account",
            "_bank_account",
            "from_bank_account",
            "_from_bank_account",
            "to_bank_account",
            "_to_bank_account",
            "currency",
            "_currency",
            "fx_rate",
            "value_local_ccy",
            "description",
            "value",
            "bank_account_currency_symbol",
        )



import factory
import pytest
from rest_framework.reverse import reverse
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.serializers import BankingContactRepresentationSerializer

from wbaccounting.models import Transaction
from wbaccounting.serializers import (
    TransactionModelSerializer,
    TransactionRepresentationSerializer,
)


@pytest.mark.django_db
class TestTransactionModelSerializer:
    def test_serialize(self, transaction: Transaction):
        # We need to add the currency symbol, which is usually done through the queryset in the viewset
        transaction.bank_account_currency_symbol = transaction.bank_account.currency.symbol  # type: ignore

        serializer = TransactionModelSerializer(transaction)
        assert transaction.currency is not None
        assert transaction.fx_rate is not None
        assert transaction.value_local_ccy is not None
        assert transaction.value is not None
        assert serializer.data == {
            "id": transaction.pk,
            "description": transaction.description,
            "booking_date": transaction.booking_date.strftime("%Y-%m-%d"),
            "value_date": transaction.value_date.strftime("%Y-%m-%d"),
            "bank_account": transaction.bank_account.id,
            "_bank_account": BankingContactRepresentationSerializer(transaction.bank_account).data,
            "from_bank_account": transaction.bank_account.id,
            "_from_bank_account": BankingContactRepresentationSerializer(transaction.bank_account).data,
            "to_bank_account": transaction.bank_account.id,
            "_to_bank_account": BankingContactRepresentationSerializer(transaction.bank_account).data,
            "currency": transaction.currency.pk,
            "_currency": CurrencyRepresentationSerializer(transaction.currency).data,
            "fx_rate": str(round(transaction.fx_rate, 4)),
            "value_local_ccy": str(round(transaction.value_local_ccy, 2)),
            "value": str(round(transaction.value, 2)),
            "bank_account_currency_symbol": transaction.bank_account.currency.symbol,
        }

    def test_deserialize(self, transaction_factory, banking_contact):
        data = factory.build(
            dict,
            FACTORY_CLASS=transaction_factory,
            bank_account=banking_contact,
        )

        data["bank_account"] = data["bank_account"].id
        data["from_bank_account"] = data["from_bank_account"].id
        data["to_bank_account"] = data["to_bank_account"].id
        data["currency"] = data["currency"].id
        data["value"] = round(data["value"], 2)

        serializer = TransactionModelSerializer(data=data)
        assert serializer.is_valid()


@pytest.mark.django_db
class TestTransactionRepresentationSerializer:
    def test_serialize(self, transaction: Transaction):
        serializer = TransactionRepresentationSerializer(transaction)
        assert transaction.value is not None
        assert serializer.data == {
            "id": transaction.pk,
            "booking_date": transaction.booking_date.strftime("%Y-%m-%d"),
            "bank_account": transaction.bank_account.id,
            "_bank_account": BankingContactRepresentationSerializer(transaction.bank_account).data,
            "value": str(round(transaction.value, 2)),
            "_detail": reverse("wbaccounting:transaction-detail", args=[transaction.pk]),
        }



from django.db.models import F, QuerySet
from wbcore import viewsets

from wbaccounting.models import Transaction
from wbaccounting.serializers import (
    TransactionModelSerializer,
    TransactionRepresentationSerializer,
)
from wbaccounting.viewsets.display import TransactionDisplayConfig


class TransactionModelViewSet(viewsets.ModelViewSet):
    queryset = Transaction.objects.all()
    serializer_class = TransactionModelSerializer
    search_fields = ("bank_account__isin", "from_bank_account__isin", "to_bank_account__isin")
    ordering = ordering_fields = ("booking_date", "value", "id")
    display_config_class = TransactionDisplayConfig

    def get_queryset(self) -> QuerySet[Transaction]:
        return (
            super()
            .get_queryset()
            .filter_for_user(user=self.request.user)
            .annotate(bank_account_currency_symbol=F("bank_account__currency__symbol"))
        )


class TransactionRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = TransactionRepresentationSerializer
    queryset = Transaction.objects.all()
    search_fields = ("bank_account__isin", "from_bank_account__isin", "to_bank_account__isin")
    ordering = ordering_fields = ("booking_date", "value", "id")

    def get_queryset(self) -> QuerySet[Transaction]:
        return super().get_queryset().filter_for_user(user=self.request.user)



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import Display, default
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class EntryAccountingInformationDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="entry", label="CRM", width=400),
                dp.Field(key="tax_id", label="Tax ID"),
                dp.Field(key="vat", label="VAT", width=100),
                dp.Field(key="default_currency", label="Currency", width=100),
                dp.Field(key="send_mail", label="Send Mail", width=100),
            ],
            formatting=[
                dp.Formatting(
                    column="counterparty_is_private",
                    formatting_rules=[dp.FormattingRule(condition=("==", True), style={"backgroundColor": "#708090"})],
                )
            ],
            legends=[
                dp.Legend(
                    key="counterparty_is_private",
                    items=[dp.LegendItem(icon="#708090", label="Private Counterparty", value=True)],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        return dp.Display(
            pages=[
                dp.Page(
                    title="General Information",
                    layouts={
                        default(): dp.Layout(
                            grid_template_areas=[
                                ["entry", "booking_entry_generator"],
                                ["counterparty_is_private", "exempt_users"],
                                ["default_currency", "tax_id"],
                                ["vat", "default_invoice_type"],
                                ["external_invoice_users", "."],
                            ],
                        )
                    },
                ),
                dp.Page(
                    title="E-Mail Settings",
                    layouts={
                        dp.default(): dp.Layout(
                            grid_template_areas=[
                                ["send_mail", "."],
                                ["email_to", "."],
                                ["email_cc", "."],
                                ["email_bcc", "."],
                                ["email_subject", "."],
                                [repeat_field(2, "email_body")],
                            ],
                            grid_template_columns=["600px", "1fr"],
                        )
                    },
                ),
                dp.Page(
                    title="Invoices",
                    layouts={
                        default(): dp.Layout(
                            grid_template_areas=[
                                ["invoices"],
                            ],
                            grid_template_rows=["600px"],
                            inlines=[dp.Inline(key="invoices", endpoint="invoices")],
                        )
                    },
                ),
                dp.Page(
                    title="Bookings",
                    layouts={
                        default(): dp.Layout(
                            grid_template_areas=[
                                ["bookingentries"],
                            ],
                            grid_template_rows=["600px"],
                            inlines=[dp.Inline(key="bookingentries", endpoint="bookingentries")],
                        )
                    },
                ),
            ]
        )



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

INVOICE_MENUITEM = MenuItem(
    label="Invoices",
    endpoint="wbaccounting:invoice-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbaccounting.view_invoice"]
    ),
    add=MenuItem(
        label="Create Invoice",
        endpoint="wbaccounting:invoice-list",
        permission=ItemPermission(permissions=["wbaccounting.add_invoice"]),
    ),
)



from django.contrib.contenttypes.models import ContentType
from django.utils.http import urlencode
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.serializers import EntryRepresentationSerializer
from wbcore.metadata.configs.buttons import WidgetButton

from wbaccounting.models import Invoice, InvoiceType
from wbaccounting.serializers import InvoiceTypeRepresentationSerializer


class InvoiceRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbaccounting:invoice-detail")

    class Meta:
        model = Invoice
        fields = (
            "id",
            "title",
            "invoice_date",
            "_detail",
        )


class InvoiceModelSerializer(serializers.ModelSerializer):
    _counterparty = EntryRepresentationSerializer(source="counterparty")
    _invoice_currency = CurrencyRepresentationSerializer(source="invoice_currency")
    _invoice_type = InvoiceTypeRepresentationSerializer(source="invoice_type")
    invoice_type = serializers.PrimaryKeyRelatedField(label="Type", required=True, queryset=InvoiceType.objects.all())

    net_value = serializers.DecimalField(max_digits=16, decimal_places=2, label="Net Value", read_only=True)
    gross_value = serializers.DecimalField(max_digits=16, decimal_places=2, label="Gross Value", read_only=True)

    @serializers.register_resource()
    def bookingentries(self, instance, request, user):
        # Do some something (checks, etc.)
        return {
            "bookingentries": reverse("wbaccounting:invoice-bookingentry-list", args=[instance.id], request=request)
        }

    @serializers.register_resource()
    def invoice_file(self, instance, request, user):
        return {
            "invoice_file": f'{reverse("wbcore:documents:document-urlredirect", args=[], request=request)}?content_type={ContentType.objects.get_for_model(Invoice).id}&object_id={instance.id}'
        }

    @serializers.register_dynamic_button()
    def dynamic_buttons(self, instance, request, user):
        buttons = []

        if backlinks := instance.backlinks:
            for _, backlink in backlinks.items():
                buttons.append(
                    WidgetButton(
                        label=backlink["title"],
                        endpoint=f'{reverse(backlink["reverse"], request=request)}?{urlencode(backlink["parameters"])}',
                    )
                )

        return buttons

    class Meta:
        model = Invoice
        decorators = {
            "net_value": serializers.decorator(
                decorator_type="text", position="left", value="{{_invoice_currency.symbol}}"
            ),
            "gross_value": serializers.decorator(
                decorator_type="text", position="left", value="{{_invoice_currency.symbol}}"
            ),
        }
        fields = (
            "id",
            "status",
            "reference_date",
            "resolved",
            "title",
            "invoice_date",
            "invoice_currency",
            "_invoice_currency",
            "counterparty",
            "_counterparty",
            "text_above",
            "text_below",
            "_additional_resources",
            "gross_value",
            "net_value",
            "invoice_type",
            "_invoice_type",
            "_buttons",
        )
        read_only_fields = (
            "gross_value",
            "net_value",
        )



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

TRANSACTION_MENUITEM = MenuItem(
    label="Transactions",
    endpoint="wbaccounting:transaction-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbaccounting.view_transaction"]
    ),
    add=MenuItem(
        label="Create Transaction",
        endpoint="wbaccounting:transaction-list",
        permission=ItemPermission(permissions=["wbaccounting.add_transaction"]),
    ),
)



from .invoice_type import InvoiceTypeModelSerializer, InvoiceTypeRepresentationSerializer
from .entry_accounting_information import (
    EntryAccountingInformationModelSerializer,
    EntryAccountingInformationRepresentationSerializer,
)
from .invoice import InvoiceRepresentationSerializer, InvoiceModelSerializer
from .booking_entry import (
    BookingEntryModelSerializer,
    BookingEntryRepresentationSerializer,
)
from .transactions import TransactionModelSerializer, TransactionRepresentationSerializer
from .consolidated_invoice import ConsolidatedInvoiceSerializer



from rest_framework.permissions import IsAuthenticated


class IsInvoiceAdmin(IsAuthenticated):
    def has_permission(self, request, view):
        return request.user.has_perm("wbaccounting.administrate_invoice")


class CanGenerateBookingEntry(IsAuthenticated):
    def has_permission(self, request, view):
        return request.user.has_perm("wbaccounting.can_generate_booking_entries")


class CanGenerateInvoice(IsAuthenticated):
    def has_permission(self, request, view):
        return request.user.has_perm("wbaccounting.can_generate_invoice")



from django.utils.http import urlencode
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.serializers import EntryRepresentationSerializer
from wbcore.metadata.configs.buttons import WidgetButton

from wbaccounting.models import BookingEntry
from wbaccounting.serializers import InvoiceRepresentationSerializer


class BookingEntryRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="wbaccounting:bookingentry-detail")

    class Meta:
        model = BookingEntry
        fields = (
            "id",
            "title",
            "_detail",
        )


class BookingEntryModelSerializer(serializers.ModelSerializer):
    _currency = CurrencyRepresentationSerializer(source="currency")
    _counterparty = EntryRepresentationSerializer(source="counterparty")
    _invoice = InvoiceRepresentationSerializer(source="invoice")

    invoice_currency = serializers.StringRelatedField(
        source="invoice.invoice_currency", read_only=True, label="Inv. Currency"
    )

    @serializers.register_dynamic_button()
    def dynamic_buttons(self, instance, request, user):
        buttons = []

        if backlinks := instance.backlinks:
            for _, backlink in backlinks.items():
                buttons.append(
                    WidgetButton(
                        label=backlink["title"],
                        endpoint=f'{reverse(backlink["reverse"], request=request)}?{urlencode(backlink["parameters"])}',
                    )
                )

        return buttons

    class Meta:
        model = BookingEntry
        decorators = {
            "gross_value": serializers.decorator(decorator_type="text", position="left", value="{{_currency.symbol}}"),
            "net_value": serializers.decorator(decorator_type="text", position="left", value="{{_currency.symbol}}"),
        }
        percent_fields = ["vat"]
        read_only_fields = ["invoice_net_value", "invoice_gross_value", "invoice_fx_rate"]

        fields = (
            "id",
            "title",
            "booking_date",
            "due_date",
            "payment_date",
            "reference_date",
            "net_value",
            "gross_value",
            "vat",
            "invoice_net_value",
            "invoice_gross_value",
            "invoice_fx_rate",
            "currency",
            "_currency",
            "counterparty",
            "_counterparty",
            "invoice",
            "_invoice",
            "invoice_currency",
            "_additional_resources",
            "_buttons",
        )



from datetime import datetime
from typing import TYPE_CHECKING

from rest_framework.reverse import reverse
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig
from wbcore.utils.models import WBColor

if TYPE_CHECKING:
    from wbaccounting.viewsets.cashflows import FutureCashFlowPandasAPIViewSetMixin


class FutureCashFlowDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        view: "FutureCashFlowPandasAPIViewSetMixin" = self.view  # type: ignore
        fields = view.get_dataframe(self.request, view.get_queryset()).columns[view.DATE_COL_START :]
        return dp.ListDisplay(
            fields=[
                dp.Field(key="bank_account__iban", label="IBAN"),
                dp.Field(key="bank_account__currency__symbol", label="Currency"),
                *[
                    dp.Field(
                        key=field,
                        label=datetime.strptime(field, "%Y-%m-%d").strftime("%d.%m.%Y"),
                        formatting_rules=[
                            dp.FormattingRule(
                                style={"color": WBColor.GREEN_DARK.value},
                                condition=(">=", 0),
                            ),
                            dp.FormattingRule(
                                style={"color": WBColor.RED_DARK.value},
                                condition=("<", 0),
                            ),
                            *[
                                dp.FormattingRule(style={"fontWeight": "bold"})
                                if view.BOLD
                                else dp.FormattingRule(style={"fontWeight": "normal"})
                            ],
                        ],
                    )
                    for field in fields
                ],
            ],
            tree=True,
            tree_group_field="bank_account__iban",
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    filter_depth=1,
                    lookup="bank_account__id",
                    filter_key="bank_account",
                    list_endpoint=reverse(
                        "wbaccounting:futurecashflowtransaction-list",
                        args=[],
                        request=self.request,
                    ),
                )
            ],
        )



from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ConsolidatedInvoiceEndpointConfig(EndpointViewConfig):
    def _get_instance_endpoint(self, **kwargs):
        return "{{casted_endpoint}}"



from datetime import date

from django.contrib import admin
from wbcore.contrib.directory.models import Entry

from wbaccounting.models import BookingEntry, Invoice


class BookingEntryInline(admin.TabularInline):
    model = BookingEntry
    fields = ("booking_date", "net_value", "vat")
    ordering = ("vat", "title")

    def has_add_permission(self, request, obj=None):
        return False


@admin.register(BookingEntry)
class BookingEntryModelAdmin(admin.ModelAdmin):
    list_display = (
        "title",
        "booking_date",
        "net_value",
        "vat",
        "currency",
        "payment_date",
        "counterparty",
        "reference_date",
    )
    search_fields = ["counterparty__computed_str"]
    autocomplete_fields = ("counterparty", "currency")

    # TODO: Remove or adjust.
    def create_invoice(self, request, queryset):
        counterparty = list(set(queryset.values_list("counterparty__id", flat=True)))
        if len(counterparty) == 1:
            counterparty = Entry.objects.get(id=counterparty[0])
            invoice = Invoice.objects.create(
                title=f"Rebate Invoice {counterparty.computed_str} ({queryset.earliest('from_date').from_date:%d.%m.%Y} - {queryset.latest('to_date').to_date:%d.%m.%Y})",
                invoice_date=date.today(),
                reference_date=date.today(),
                counterparty=counterparty,
                invoice_currency=counterparty.entry_accounting_information.default_currency,
                is_counterparty_invoice=True,
            )
            queryset.update(invoice=invoice)
            invoice.save()

        else:
            print("-------------------------------")  # noqa: T201
            print(f"Too many counterparties selected ({len(counterparty)})")  # noqa: T201
            print("-------------------------------")  # noqa: T201

    actions = [create_invoice]



from .booking_entry import BookingEntryDisplayConfig
from .invoice import InvoiceDisplayConfig, ConsolidatedInvoiceDisplayConfig
from .invoice_type import InvoiceTypeDisplayConfig
from .entry_accounting_information import (
    EntryAccountingInformationDisplayConfig,
)
from .transactions import TransactionDisplayConfig
from .cashflows import FutureCashFlowDisplayConfig



import hashlib
from contextlib import suppress
from datetime import datetime

import pandas as pd
from wbportfolio.models import Product


def parse(import_source):
    csv_file = import_source.file.open()
    df = pd.read_csv(csv_file)
    df = df[~df["Isin"].isnull()]
    df["Trade date"] = df["Trade date"].ffill()

    df_sub = df[
        ["Trade date", "Isin", "Quantity Sub", "Amount Sub class ccy", "Value date Sub", "Nav Class ccy", "Ccy Class"]
    ]
    df_sub = df_sub.rename(
        columns={"Quantity Sub": "quantity", "Amount Sub class ccy": "amount", "Value date Sub": "value_date"}
    )
    df_red = df[
        ["Trade date", "Isin", "Quantity Red", "Amount Red Class ccy", "Value date Red", "Nav Class ccy", "Ccy Class"]
    ]
    df_red = df_red.rename(
        columns={"Quantity Red": "quantity", "Amount Red Class ccy": "amount", "Value date Red": "value_date"}
    )
    df_red["quantity"] = df_red["quantity"] * -1
    df_red["amount"] = df_red["amount"] * -1
    df = pd.concat([df_sub, df_red])
    df["value"] = df["amount"] + (df["quantity"] * df["Nav Class ccy"])

    df = df[df["value"] != 0]

    # Iterate through the CSV File and parse the data into a list
    data = list()
    for transaction in df.to_dict(orient="records"):
        _hash = hashlib.sha256()
        for field in transaction.values():
            _hash.update(str(field).encode())

        description = f'Trade ({transaction["Isin"]}): {transaction["quantity"]} shares and {transaction["amount"]} {transaction["Ccy Class"]}.'

        with suppress(Product.DoesNotExist):
            bank_account = Product.objects.get(isin=transaction["Isin"]).bank_account.iban
            data.append(
                {
                    "booking_date": datetime.strptime(transaction["Trade date"], "%d/%m/%Y").strftime("%Y-%m-%d"),
                    "value_date": datetime.strptime(transaction["value_date"], "%d/%m/%Y").strftime("%Y-%m-%d"),
                    "currency": transaction.get("Ccy Class"),
                    "value": float(transaction["value"]),
                    "description": description,
                    "bank_account": bank_account,
                    "prenotification": True,
                    "_hash": _hash.hexdigest(),
                }
            )

    return {
        "data": data,
    }



from django.utils.http import urlencode
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.serializers import EntryRepresentationSerializer

from wbaccounting.models import BookingEntry, Invoice
from wbaccounting.serializers import (
    BookingEntryRepresentationSerializer,
    InvoiceRepresentationSerializer,
    InvoiceTypeRepresentationSerializer,
)


class ConsolidatedInvoiceSerializer(serializers.Serializer):
    id = serializers.PrimaryKeyCharField()
    reference_date = serializers.DateField()
    invoice = serializers.PrimaryKeyRelatedField(allow_null=True)
    _invoice = InvoiceRepresentationSerializer(source="invoice")
    invoice_currency = serializers.PrimaryKeyRelatedField(allow_null=True)
    _invoice_currency = CurrencyRepresentationSerializer(source="invoice_currency")
    counterparty = serializers.PrimaryKeyRelatedField(allow_null=True)
    _counterparty = EntryRepresentationSerializer(source="counterparty")
    _group_key = serializers.CharField()
    type = serializers.PrimaryKeyRelatedField(allow_null=True)
    _type = InvoiceTypeRepresentationSerializer(source="type")
    booking_entries = serializers.PrimaryKeyRelatedField(allow_null=True)
    _booking_entries = BookingEntryRepresentationSerializer(source="booking_entries")
    group = serializers.CharField()
    currency_symbol = serializers.CharField()
    value = serializers.DecimalField(
        max_digits=15,
        decimal_places=2,
        decorators=[serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")],
    )
    casted_endpoint = serializers.SerializerMethodField()
    depth = serializers.IntegerField()
    num_draft = serializers.IntegerField()
    num_submitted = serializers.IntegerField()
    num_sent = serializers.IntegerField()
    num_paid = serializers.IntegerField()

    @serializers.register_resource()
    def register_buttons(self, instance: dict, request, user: User) -> dict[str, str]:
        button_dict = {}

        if instance["depth"] == 5:
            return button_dict

        if instance.get("num_draft", 0) > 0:
            button_dict["submit"] = reverse(
                "wbaccounting:consolidated-invoice-submit", args=[instance["id"]], request=request
            )

        if instance.get("num_submitted", 0) > 0 and user.has_perm("wbaccounting.administrate_invoice"):
            button_dict["approve"] = reverse(
                "wbaccounting:consolidated-invoice-approve", args=[instance["id"]], request=request
            )

        if instance.get("num_sent", 0) > 0 and user.has_perm("wbaccounting.administrate_invoice"):
            button_dict["pay"] = reverse(
                "wbaccounting:consolidated-invoice-pay", args=[instance["id"]], request=request
            )

        for key in button_dict.keys():
            button_dict[key] = button_dict[key] + "?" + urlencode(dict(request.GET.items()))
        return button_dict

    def get_casted_endpoint(self, obj: dict) -> str | None:
        if obj.get("depth", 0) == 4:
            return reverse(
                f"{Invoice.get_endpoint_basename()}-detail",
                args=[obj["id"]],
                request=self.context.get("request", None),
            )
        elif obj.get("depth", 0) == 5:
            return reverse(
                f"{BookingEntry.get_endpoint_basename()}-detail",
                args=[obj["id"]],
                request=self.context.get("request", None),
            )

    class Meta:
        fields = read_only_fields = (
            "id",
            "reference_date",
            "invoice",
            "_invoice",
            "invoice_currency",
            "_invoice_currency",
            "counterparty",
            "_counterparty",
            "booking_entries",
            "_booking_entries",
            "group",
            "value",
            "_additional_resources",
            "type",
            "_type",
            "casted_endpoint",
            "currency_symbol",
            "_buttons",
            "_group_key",
            "depth",
            "num_draft",
            "num_submitted",
            "num_sent",
            "num_paid",
        )



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

ENTRYACCOUNTINGINFORMATION_MENUITEM = MenuItem(
    label="Counterparties",
    endpoint="wbaccounting:entryaccountinginformation-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbaccounting.view_entryaccountinginformation"],
    ),
)



from contextlib import suppress
from datetime import date

from celery import shared_task
from django.db.models.signals import post_save
from django.utils.module_loading import import_string
from wbcore.contrib.directory.models import Entry

from wbaccounting.generators.base import generate_booking_entries


@shared_task
def submit_invoices_as_task(ids: list[int]):
    from wbaccounting.models import Invoice

    invoices = Invoice.objects.filter(id__in=ids)
    for invoice in invoices:
        invoice.submit()
        invoice.save()


@shared_task
def approve_invoices_as_task(ids: list[int]):
    from wbaccounting.models import Invoice

    invoices = Invoice.objects.filter(id__in=ids)
    for invoice in invoices:
        invoice.approve()
        invoice.save()


@shared_task
def pay_invoices_as_task(ids: list[int]):
    from wbaccounting.models import Invoice

    invoices = Invoice.objects.filter(id__in=ids)
    for invoice in invoices:
        invoice.pay()
        invoice.save()


@shared_task
def refresh_complete_invoice_as_task(invoice_id: int):
    from wbaccounting.models import Invoice
    from wbaccounting.models.booking_entry import BookingEntry, booking_entry_changed

    # We temmporarily disconnect the post_save hook in order to not regenerate the invoice over and over again
    post_save.disconnect(booking_entry_changed, sender=BookingEntry)

    for booking_entry in BookingEntry.objects.filter(invoice_id=invoice_id):
        booking_entry.save()

    Invoice.objects.get(id=invoice_id).save()

    # We reattach the post_save hook
    post_save.connect(booking_entry_changed, sender=BookingEntry)


@shared_task
def refresh_invoice_document_as_task(invoice_id):
    from wbaccounting.models import Invoice

    invoice = Invoice.objects.get(id=invoice_id)
    invoice.refresh_invoice_document()


@shared_task
def generate_booking_entries_as_task(func: str, from_date: date, to_date: date, counterparty_id: int):
    with suppress(ImportError):
        generator = import_string(func)
        counterparty = Entry.objects.get(id=counterparty_id)
        generate_booking_entries(generator, from_date, to_date, counterparty)



from django.contrib import admin
from wbcore.contrib.documents.admin import DocumentInLine

from wbaccounting.admin import BookingEntryInline
from wbaccounting.models import Invoice


@admin.register(Invoice)
class InvoiceModelAdmin(admin.ModelAdmin):
    fsm_field = ["status"]
    search_fields = ("counterparty__computed_str", "title", "invoice_type__name")
    list_display = (
        "status",
        "title",
        "gross_value",
        "net_value",
        "invoice_date",
        "invoice_currency",
        "counterparty",
        "invoice_type",
        "reference_date",
    )

    autocomplete_fields = ("counterparty",)
    inlines = [BookingEntryInline, DocumentInLine]

    raw_id_fields = ["counterparty", "invoice_currency", "invoice_type"]



from .booking_entry import BookingEntry
from .invoice import Invoice
from .invoice_type import InvoiceType
from .entry_accounting_information import EntryAccountingInformation
from .transactions import Transaction
from .model_tasks import submit_invoices_as_task



import pytest
from pytest_factoryboy.fixture import LazyFixture
from wbcore.contrib.authentication.models import User

from wbaccounting.models import EntryAccountingInformation, Invoice


@pytest.mark.django_db
class TestInvoice:
    def test_str(self, invoice: Invoice):
        assert str(invoice) == invoice.title

    @pytest.mark.parametrize(
        "method,return_value",
        [
            ("get_endpoint_basename", "wbaccounting:invoice"),
            ("get_representation_value_key", "id"),
            ("get_representation_label_key", "{{title}} ({{invoice_date}})"),
            ("get_representation_endpoint", "wbaccounting:invoicerepresentation-list"),
        ],
    )
    def test_wbmodel_methods(self, method: str, return_value: str):
        assert getattr(Invoice, method)() == return_value

    @pytest.mark.parametrize("user__is_superuser", [False])
    def test_filter_for_user_no_superuser(self, invoice: Invoice, user: User):
        invoices = Invoice.objects.filter_for_user(user)  # type: ignore
        assert invoice not in invoices

    @pytest.mark.parametrize("user__is_superuser", [True])
    def test_filter_for_user_superuser(self, invoice: Invoice, user: User):
        invoices = Invoice.objects.filter_for_user(user)  # type: ignore
        assert invoice in invoices

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_invoice"])])
    @pytest.mark.parametrize("invoice__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [False])
    def test_filter_for_user_public_counterparty(
        self, invoice: Invoice, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        invoices = Invoice.objects.filter_for_user(user)  # type: ignore
        assert invoice in invoices

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("invoice__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty(
        self, user: User, invoice: Invoice, entry_accounting_information: EntryAccountingInformation
    ):
        invoices = Invoice.objects.filter_for_user(user)  # type: ignore
        assert invoice not in invoices

    @pytest.mark.parametrize("user__is_superuser", [False])
    @pytest.mark.parametrize("user__user_permissions", [(["wbaccounting.view_entryaccountinginformation"])])
    @pytest.mark.parametrize("invoice__counterparty", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__entry", [LazyFixture("entry")])
    @pytest.mark.parametrize("entry_accounting_information__counterparty_is_private", [True])
    def test_filter_for_user_private_counterparty_with_exempt(
        self, invoice: Invoice, user: User, entry_accounting_information: EntryAccountingInformation
    ):
        entry_accounting_information.exempt_users.add(user)
        entry_accounting_information_list = EntryAccountingInformation.objects.filter_for_user(user)  # type: ignore
        assert entry_accounting_information in entry_accounting_information_list



import logging
from contextlib import suppress
from io import BytesIO

from dynamic_preferences.registries import global_preferences_registry
from reportlab import platypus
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import cm
from reportlab.platypus import BaseDocTemplate, NextPageTemplate, PageTemplate
from reportlab.platypus.frames import Frame
from wbcore.contrib.directory.models import Entry

from wbaccounting.files import utils as invoice_utils

logger = logging.getLogger("wbaccounting.files.invoice_document_file")


def generate_file(invoice):
    debug = 0

    total_gross_value, total_net_value = invoice_utils.get_gross_and_net_value(invoice)
    show_tax_id = total_gross_value != total_net_value

    styles = invoice_utils.get_styles()

    output = BytesIO()
    doc = BaseDocTemplate(output, pagesize=A4, rightMargin=10, leftMargin=10, topMargin=10, bottomMargin=10)

    global_preferences = global_preferences_registry.manager()
    invoice_company = Entry.objects.get(id=global_preferences["wbaccounting__invoice_company"])
    signees = global_preferences["wbaccounting__invoice_signers"]

    logo_ratio = 0
    logo = None
    try:
        logo, logo_ratio = invoice_utils.logo_block(invoice_company.profile_image)
    except ValueError:
        pass
    logo_frame = Frame(
        x1=0,
        y1=A4[1] - (30 + 1 * cm),
        width=30 * logo_ratio + 1.5 * cm,
        height=30 + 1 * cm,
        leftPadding=1.5 * cm,
        bottomPadding=0,
        rightPadding=0,
        topPadding=1 * cm,
        id="logo_frame",
        showBoundary=debug,
    )

    address_frame = Frame(
        x1=A4[0] - (12 * cm),
        y1=A4[1] - (6 * cm),
        width=12 * cm,
        height=6 * cm,
        leftPadding=0,
        bottomPadding=0,
        rightPadding=1.5 * cm,
        topPadding=2.5 * cm,
        id="address_frame",
        showBoundary=debug,
    )

    content_frame = Frame(
        x1=0,
        y1=0,
        width=A4[0],
        height=A4[1] - (6 * cm),
        leftPadding=1.5 * cm,
        bottomPadding=1 * cm,
        rightPadding=1.5 * cm,
        topPadding=0,
        id="content_frame",
        showBoundary=debug,
    )

    content_frame2 = Frame(
        x1=0,
        y1=0,
        width=A4[0],
        height=A4[1],
        leftPadding=1.5 * cm,
        bottomPadding=1 * cm,
        rightPadding=1.5 * cm,
        topPadding=1 * cm,
        id="content_frame2",
        showBoundary=debug,
    )

    doc.addPageTemplates(
        [
            PageTemplate(id="TitlePage", frames=[logo_frame, address_frame, content_frame]),
            PageTemplate(id="OtherPage", frames=[content_frame2]),
        ]
    )

    elements = list()
    elements.append(NextPageTemplate(["OtherPage"]))
    if logo:
        elements.append(logo)
    elements.append(platypus.FrameBreak("address_frame"))

    if invoice.is_counterparty_invoice:
        elements.extend(invoice_utils.address_block(invoice_company, invoice, styles, right=True))
    else:
        elements.extend(invoice_utils.address_block(invoice.counterparty, invoice, styles, right=True))

    elements.append(platypus.FrameBreak("content_frame"))
    elements.extend(invoice_utils.city_and_date_block(invoice, styles, entry=invoice_company))

    if invoice.is_counterparty_invoice:
        elements.extend(invoice_utils.address_block(invoice.counterparty, invoice, styles, tax_id=show_tax_id))
    else:
        elements.extend(invoice_utils.address_block(invoice_company, invoice, styles, tax_id=show_tax_id))

    elements.extend(invoice_utils.add_text_block_with_context(invoice.text_above, invoice, styles))
    elements.extend(invoice_utils.add_booking_entries(invoice, styles, False))
    elements.extend(invoice_utils.add_text_block_with_context(invoice.text_below, invoice, styles))
    elements.extend(
        invoice_utils.add_transfer_block(
            invoice,
            invoice_company if not invoice.is_counterparty_invoice else invoice.counterparty,
            styles,
            total_gross_value,
        )
    )
    with suppress(ValueError):
        elements.extend(invoice_utils.signature_block(signees, styles))
    doc.build(elements)

    pdf = output.getvalue()
    output.close()

    return pdf



from decimal import Decimal

import factory

from wbaccounting.models import Transaction


class AbstractTransactionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Transaction
        abstract = True

    booking_date = factory.Faker("date_between", start_date="+2d", end_date="+3d")
    value_date = factory.Faker("date_object")

    bank_account = factory.SubFactory("wbcore.contrib.directory.factories.BankingContactFactory")
    from_bank_account = factory.SubFactory("wbcore.contrib.directory.factories.BankingContactFactory")
    to_bank_account = factory.SubFactory("wbcore.contrib.directory.factories.BankingContactFactory")

    @factory.post_generation
    def set_currency(self, create, extracted, **kwargs):
        if isinstance(self, dict):
            self["currency"] = self["bank_account"].currency
        else:
            self.currency = self.bank_account.currency


class TransactionFactory(AbstractTransactionFactory):
    value = factory.Faker("pydecimal", min_value=Decimal(0.1), max_value=Decimal(10000000))


class LocalCurrencyTransactionFactory(AbstractTransactionFactory):
    value_local_ccy = factory.Faker("pydecimal", min_value=Decimal(0.1), max_value=Decimal(10000000))



from decimal import Decimal
from typing import Any

from django.db import models
from django.db.models import Value
from django.db.models.functions import Replace
from wbcore.contrib.currency.import_export.handlers import CurrencyImportHandler
from wbcore.contrib.directory.models import BankingContact
from wbcore.contrib.io.imports import ImportExportHandler


class TransactionImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbaccounting.Transaction"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.currency_handler = CurrencyImportHandler(self.import_source)

    def _deserialize(self, data: dict[str, Any]):
        data["currency"] = self.currency_handler.process_object({"key": data["currency"]}, read_only=True)[0]
        data["bank_account"] = BankingContact.objects.annotate(
            stripped_iban=Replace("iban", Value(" "), Value(""))
        ).get(stripped_iban=data["bank_account"].replace(" ", ""))
        data["value"] = Decimal(data["value"])
        return super()._deserialize(data)

    def _get_instance(self, data: dict[str, Any], history: models.QuerySet | None = None, **kwargs) -> Any | None:
        if _hash := data.get("_hash", None):
            qs = self.model.objects.filter(_hash=_hash)
            if qs.exists():
                return qs.first()
        return super()._get_instance(data, history, **kwargs)
